# inspired by http://blog.fairchild.dk/2013/06/automagical-asm-control-flow-arrow-annotation/

oo::class create ::tcl::unsupported::Disassembler {
    variable charmap colormap
    constructor {} {
	set charmap {
	    {} " "
	    {d u} "\u2502"
	    {l r} "\u2500"
	    {r u} "\u2514"
	    {d r} "\u250c"
	    {d l r u} "\u253c"
	    {l r x} "\u25ba"
	    {l r u} "\u2534"
	    {d l r} "\u252c"
	    {jump} "\u27a1"
	}
	set colormap {
	    1,1 "\u001b\[33m"
	    1,0 "\u001b\[32m"
	    0,1 "\u001b\[31m"
	    0,0 "\u001b\[35m"
	    break "\u001b\[34m"
	    continue "\u001b\[31m"
	    highlight "\u001b\[37m"
	    reset "\u001b\[m"
	}
    }

    method FmtVarRef {var} {
	lassign [lindex [dict get $dis variables] $var] flags name
	return %[if {"temp" ni $flags} {subst {$name}} else {subst {%%$var}}]
    }
    method FmtJump {to {base 0}} {
	upvar 1 from from jumps jumps
	incr to $base	
	lappend jumps [list $from $to [llength $jumps]]
	return "[dict get $charmap jump] $to"
    }
    method FmtLiteral {index} {
	set lit [lindex [dict get $dis literals] $index]
	# Oooooh! Evil!
	regsub "^\\\\?\\\{" [list \{$lit] "" lit
	return "\"$lit\""
    }

    method ParseInstructions {} {
	set jumps {}
	set prefix {{} {}}
	dict for {from inst} [dict get $dis instructions] {
	    set idx 0
	    foreach argument [lrange $inst 1 end] {
		incr idx
		if {[regexp {^pc (\d+)$} $argument -> to]} {
		    lset inst $idx [my FmtJump $to]
		    lappend prefix {}
		} elseif {[regexp {^\?(\d+)$} $argument -> auxIdx]} {
		    set aux [lindex [dict get $dis auxiliary] $auxIdx]
		    switch [dict get $aux name] {
			"JumptableInfo" {
			    # Doesn't scale very well...
			    dict for {str offset} [dict get $aux mapping] {
				my FmtJump $offset $from
				lappend prefix {}
			    }
			    lset inst $idx ""
			}
			"ForeachInfo" {
			    set aux [dict remove $aux name]
			    dict with aux {
				set i 0
				foreach var $data {
				    lset data $i [my FmtVarRef $var]
				    incr i
				}
				set i 0
				foreach var $loop {
				    lset loop $i [my FmtVarRef $var]
				    incr i
				}
				for {set i 0} {$i < [llength $assign]} {incr i} {
				    set j 0
				    foreach var [lindex $assign $i] {
					lset assign $i $j [my FmtVarRef $var]
					incr j
				    }
				}
			    }
			    lset inst $idx [list $aux]
			}
			"DictUpdateInfo" {
			    lset inst $idx [list [lmap var [dict get $aux variables] {
				my FmtVarRef $var
			    }]]
			}
			default {
			    # Fallback; shouldn't happen for now
			    lset inst $idx [list $aux]
			}
		    }
		} elseif {[regexp {^@(\d+)$} $argument -> lit]} {
		    lset inst $idx [my FmtLiteral $lit]
		} elseif {[regexp {^%(\d+)$} $argument -> var]} {
		    lset inst $idx [my FmtVarRef $var]
		}
	    }
	    dict set dis instructions $from $inst
	}
	set alen [string length $from]

	return [list $jumps $prefix $alen]
    }

    method RenderInstructionLine {
	address maxAddressLength catchRangeCount loopRangeCount backjumpCount
	targets instruction prefix
    } {

	set reset [dict get $colormap reset]
	if {"catch" in $targets} {
	    set acol [dict get $colormap 1,1]
	} elseif {"break" in $targets} {
	    set acol [dict get $colormap break]
	} elseif {"continue" in $targets} {
	    set acol [dict get $colormap continue]
	} else {
	    set acol $reset
	}
	set color [dict get $colormap [expr {
	    $catchRangeCount & 1
	}],[expr {
	    $loopRangeCount & 1
	}]]
	set prefixStr [join [lmap n $prefix {dict get $charmap $n}] ""]
	set indent [expr {$backjumpCount * 2}]
	puts [format "%s %s%*d%s %*s%s%s%s" \
		$prefixStr $acol $maxAddressLength $address $reset \
		$indent "" $color [join $instruction] $reset]
    }

    method PrintProgram {string} {
	foreach line [split $string \n] {
	    puts [dict get $colormap highlight]$line[dict get $colormap reset]
	}
    }
    method ResetTerminal {} {
	puts -nonewline [dict get $colormap reset]
	flush stdout
    }

    method OptimizeJumpCollection {} {
	upvar 1 jumps jumps prefix prefix
	# Optimize the apparent depths of non-overlapping jumps/loops
	for {set pi -1;set i 0} {[set j $i] < [llength $jumps]} {incr i} {
	    lassign [lindex $jumps $i] from to idx
	    if {$idx <= $pi} continue {set pi $idx}
	    set max [expr {max($from, $to)}]
	    while {[incr j] < [llength $jumps]} {
		lassign [lindex $jumps $j] f2 t2 i2
		if {$i2 > $idx && $max < min($f2, $t2)} {
		    lset jumps $j 2 $idx
		    lassign [lindex $jumps $j] from to idx
		    set max [expr {max($from, $to)}]
		}
	    }
	}
	set indices [lsort -integer -unique [lmap j $jumps {lindex $j 2}]]
	for {set i [llength $jumps]} {[incr i -1] >= 0} {} {
	    if {$i in $indices} continue
	    for {set j 0} {$j < [llength $jumps]} {incr j} {
		if {[lindex $jumps $j 2] > $i} {
		    lset jumps $j 2 [expr {[lindex $jumps $j 2] - 1}]
		}
	    }
	    set prefix [lrange $prefix 0 end-1]
	}
    }

    method DecodeJumpPattern {outl len addr jump} {
	upvar 1 j jumpCount
	lassign $jump from to i
	# Reverse the order; looks better
	set i [expr {$len - $i - 2}]
	if {$from == $addr} {
	    lset outl $i end+1 [lindex "u d" [expr {$from<$to}]]
	    lset outl $i end+1 "r"
	    while {[incr i] < $len} {
		lset outl $i end+1 "l"
		lset outl $i end+1 "r"
	    }
	} elseif {$to == $addr} {
	    lset outl $i end+1 [lindex "u d" [expr {$from>$to}]]
	    lset outl $i end+1 "r"
	    while {[incr i] < $len} {
		lset outl $i end+1 "l"
		lset outl $i end+1 "r"
	    }
	    lset outl [incr i -1] end+1 "x"
	} elseif {$from<$addr && $to>$addr || $from>$addr && $to<$addr} {
	    lset outl $i end+1 "u"
	    lset outl $i end+1 "d"
	}
	if {$from>$addr && $to<=$addr} {
	    incr jumpCount
	}
	return $outl
    }

    method GetExceptionRangeInfo {exns addr} {
	set c [set l 0]
	set targets {}
	foreach exn $exns {
	    if {$addr>=[dict get $exn from] && $addr<=[dict get $exn to]} {
		switch [dict get $exn type] {
		    loop {incr l}
		    catch {incr c}
		}
	    }
	    if {[dict exists $exn catch] && [dict get $exn catch] == $addr} {
		lappend targets catch
	    }
	    if {[dict exists $exn break] && [dict get $exn break] == $addr} {
		lappend targets break
	    }
	    if {[dict exists $exn continue] && [dict get $exn continue] == $addr} {
		lappend targets continue
	    }
	}
	return [list $c $l $targets]
    }

    method Callback {methodName args} {
	list [namespace current]::my $methodName {*}$args
    }
    forward GetBytecode ::tcl::unsupported::getbytecode

    variable dis
    method Disassemble {type args} {
	# Pass the arguments through to the underlying bytecode retrieval
	# system...
	set dis [uplevel 1 [my Callback GetBytecode $type {*}$args]]

	# Build the collection of jumps
	lassign [my ParseInstructions] jumps prefix alen

	# Optimize the apparent depths of non-overlapping jumps/loops
	my OptimizeJumpCollection

	my PrintProgram [lreplace [info level 0] 0 1 [self]]
	# Print the output
	set exns [dict get $dis exception]
	try {
	    dict for {addr inst} [dict get $dis instructions] {
		set pat $prefix
		set j 0
		foreach jump $jumps {
		    set pat [my DecodeJumpPattern $pat [llength $prefix] $addr $jump]
		}

		lassign [my GetExceptionRangeInfo $exns $addr] c l targets

		my RenderInstructionLine $addr $alen $c $l $j $targets $inst \
		    [lmap p $pat {lsort -unique $p}]
	    }
	} finally {
	    my ResetTerminal
	}
    }

    # Do it this way to get best error messages
    method proc {name} {
	tailcall my Disassemble proc $name
    }
    method script {script} {
	tailcall my Disassemble script $script
    }
    method lambda {lambdaTerm} {
	tailcall my Disassemble lambda $lambdaTerm
    }
    method method {class method} {
	tailcall my Disassemble method $class $method
    }
    method objmethod {object method} {
	tailcall my Disassemble objmethod $object $method
    }
    unexport destroy
}

tcl::unsupported::Disassembler create tcl::unsupported::controlflow

if {$::argv0 eq [info script]} {
    catch {clock format}
    apply {{} {
	foreach m [info class methods Disassembler -private] {
	    if {[info class methodtype Disassembler $m] eq "method"} {
		controlflow method Disassembler $m
	    }
	}
	controlflow proc tcl::clock::ParseClockFormatFormat
    } ::tcl::unsupported}
}
